// 013 u_c_find_uses
// 03 jan 2002

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_find_uses;
  interface
    uses Classes
        , u_c_text_file
        , u_c_uses_list
        ;

    type c_find_uses= class(c_text_file)
                      public
                        constructor create_find_uses(p_name, p_file_name: String); Virtual;
                        procedure parse(p_unit_name: String; p_level: Integer; p_c_uses_list: c_uses_list);
                        destructor Destroy; Override;
                    end;

    function f_is_custom_unit(p_name: String): Boolean;
    function f_is_project(p_name: String): Boolean;
    function f_is_dpr(p_name: String): Boolean;

  implementation
    uses SysUtils, Dialogs, FileCtrl
      , u_types_constants
      , u_c_log, u_c_display
      , u_c_file_name
      ;

    function f_is_custom_unit(p_name: String): Boolean;
      var l_name: String;
      begin
        l_name:= ExtractFileName(p_name);
        Result:= (Length(l_name)> 0) and (l_name[1]= 'u');
      end; // f_is_custom_unit

    function f_is_dpr(p_name: String): Boolean;
      var l_extension: String;
      begin
        l_extension:= ExtractFileExt(p_name);
        Result:= (UpperCase(ExtractFileExt(p_name))= '.DPR');
      end; // f_is_dpr

    function f_is_project(p_name: String): Boolean;
      begin
        Result:= (Length(p_name)> 0) and (p_name[1]= 'p');
      end; // f_is_project

    // -- c_find_uses

    constructor c_find_uses.create_find_uses(p_name, p_file_name: String);
      begin
        Inherited create_text_file(p_name, p_file_name);
      end; // create_find_uses

    procedure c_find_uses.parse(p_unit_name: String; p_level: Integer; p_c_uses_list: c_uses_list);
      // var l_result_line: String;

      procedure error(p_text: String);
        begin
          ShowMessage(IntToStr(m_buffer_index)+ ' '+ p_text);
        end; // error

      procedure analyze;
        type t_symbol_type= (e_unknown,
                            e_identifier, e_number, e_operator, e_string_litteral,
                            e_comment, e_comment_slash,
                            e_end_symbol);
        var l_blank_string, l_symbol_string: String;
            l_symbol_type: t_symbol_type;

        function f_symbol_type: String;
          begin
            case l_symbol_type of
              e_unknown : Result:= 'e_unknown';
              e_identifier: Result:= 'e_identifier';
              e_number: Result:= 'e_number';
              e_operator: Result:= 'e_operator';
              e_string_litteral: Result:= 'e_string_litteral';
              e_comment: Result:= 'e_comment';
              e_comment_slash: Result:= 'e_comment_slash';
              e_end_symbol: Result:= 'e_end_symbol';
              else Result:= '??';
            end;
          end; // f_symbol_type

        procedure read_symbol;
          const k_blanks= [' ', chr(9)];
                k_returns= [k_return, k_line_feed];
                k_blanks_returns= k_blanks+ k_returns;
                k_number= ['0'..'9', '.'];
                k_identifier= ['a'..'z', 'A'..'Z', '_', '0'..'9'];
                k_other= k_blanks_returns+ k_identifier;

          procedure get_blanks;
            var l_start, l_length: Integer;
            begin
              l_start:= m_buffer_index;
              while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_blanks_returns) do
                inc(m_buffer_index);

              l_length:= m_buffer_index- l_start;
              if l_length> 0
                then begin
                    SetLength(l_blank_string, l_length);
                    Move(m_pt_buffer[l_start], l_blank_string[1], l_length);
                  end
                else l_blank_string:= '';
            end; // get_blanks

          procedure get_identifier;
            var l_start, l_length: Integer;
            begin
              l_symbol_type:= e_identifier;
              l_start:= m_buffer_index;
              while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_identifier) do
                inc(m_buffer_index);
              l_length:= m_buffer_index- l_start;
              if l_length> 0
                then begin
                    SetLength(l_symbol_string, l_length);
                    Move(m_pt_buffer[l_start], l_symbol_string[1], l_length);
                  end
                else l_symbol_string:= '';
            end; // get_identifier

          procedure get_number;
            var l_start, l_length: Integer;
            begin
              l_symbol_type:= e_number;
              l_start:= m_buffer_index;
              while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_number) do
                inc(m_buffer_index);
              l_length:= m_buffer_index- l_start;
              if l_length> 0
                then begin
                    SetLength(l_symbol_string, l_length);
                    Move(m_pt_buffer[l_start], l_symbol_string[1], l_length);
                  end
                else l_symbol_string:= '';
            end; // get_number

          procedure get_string_litteral;
            var l_start, l_length: Integer;
            begin
              l_symbol_type:= e_string_litteral;
              l_start:= m_buffer_index;

              repeat
                repeat
                  inc(m_buffer_index);
                until m_pt_buffer[m_buffer_index]= '''';

                // -- get rid of double ''
                inc(m_buffer_index);
              until m_pt_buffer[m_buffer_index]<> '''';

              l_length:= m_buffer_index- l_start;
              l_symbol_string:= f_extract_string_start_length(l_start, l_length);
            end; // get_string_litteral

          procedure get_operator;
              // -- get 1 character
            var l_start, l_length: Integer;
            begin
              l_symbol_type:= e_operator;
              l_start:= m_buffer_index;
              Inc(m_buffer_index);
              l_length:= m_buffer_index- l_start;
              if l_length> 0
                then begin
                    SetLength(l_symbol_string, l_length);
                    Move(m_pt_buffer[l_start], l_symbol_string[1], l_length);
                  end
                else l_symbol_string:= '';
            end; // get_operator

          procedure skip_comment_parenthesis;
            begin
              l_symbol_type:= e_comment;
              l_symbol_string:= '';
              Inc(m_buffer_index, 2);

              while (m_buffer_index+ 1< m_buffer_size)
                  and
                      (
                         (m_pt_buffer[m_buffer_index]<> '*')
                        or
                          (m_pt_buffer[m_buffer_index+ 1]<> ')')
                      )
                       do
                inc(m_buffer_index);

              Inc(m_buffer_index, 2);
            end; // skip_comment_parenthesis

          procedure skip_comment_slash;
            begin
              l_symbol_type:= e_comment_slash;

              l_symbol_string:= '';
              Inc(m_buffer_index, 2);
              while (m_buffer_index< m_buffer_size)
                  and (m_pt_buffer[m_buffer_index]<> k_return)
                  and (m_pt_buffer[m_buffer_index]<> k_line_feed) do
                inc(m_buffer_index);
            end; // skip_comment_slash

          procedure skip_comment_curly_braket;
            begin
              l_symbol_type:= e_comment;
              l_symbol_string:= '';
              Inc(m_buffer_index, 1);
              while (m_buffer_index< m_buffer_size)
                  and (m_pt_buffer[m_buffer_index]<> '}') do
                inc(m_buffer_index);
              Inc(m_buffer_index);
            end; // skip_comment_curly_braket

          begin // read_symbol
            get_blanks;

            if m_buffer_index< m_buffer_size
              then begin
                  case m_pt_buffer[m_buffer_index] of
                    'a'..'z', 'A'..'Z' : get_identifier;
                    '0'..'9' : get_number;
                    '''' : get_string_litteral;
                    '(' : if (m_buffer_index+ 1< m_buffer_size) and (m_pt_buffer[m_buffer_index+ 1]= '*')
                            then skip_comment_parenthesis
                            else get_operator;
                    '/' : if (m_buffer_index+ 1< m_buffer_size) and (m_pt_buffer[m_buffer_index+ 1]= '/')
                            then skip_comment_slash
                            else get_operator;
                    '{' : skip_comment_curly_braket;
                    else get_operator;
                  end; // case
                end
              else l_symbol_type:= e_end_symbol;
          end; // read_symbol

        // -- get the id's

        procedure skip_until_id(p_id: String);
          begin
            while not f_end_of_text do
            begin
              if (l_symbol_type= e_identifier) and (UpperCase(l_symbol_string)= p_id)
                then break;
              read_symbol;
            end; // while not f_end_of_text
          end; // skip_until_id

        function f_next_id: String;
          begin
            while not f_end_of_text do
            begin
              read_symbol;
              if l_symbol_type= e_identifier
                then break;
            end; // while not f_end_of_text

            Result:= l_symbol_string;
          end; // f_next_id

        // -- analyze specific parts

        procedure add_a_file_name(p_the_unit_name: String);
          begin
            if f_is_custom_unit(p_the_unit_name) or f_is_project(p_the_unit_name)
              then
                p_c_uses_list.add_new_use_or_increment_count(p_the_unit_name,
                    ExtractFilePath(p_unit_name), p_level);
          end; // add_a_file_name

        procedure analyze_project;

          procedure analyze_project_name;
            var l_project_name: String;
            begin // analyze_unit_name
              display('> analyze_project_name');

              // skip PROGRAM
              f_next_id;

              l_project_name:= f_next_id;

              // skip ;
              read_symbol;

              display('< analyze_project_name '+ l_project_name+ '<');
            end; // analyze_project_name

          procedure analyze_project_uses_list;

            procedure analyse_the_uses_list;
              begin
                while not f_end_of_text and (l_symbol_string<> ';') do
                begin
                  read_symbol;
                  if (l_symbol_type= e_identifier)
                      and (Length(l_symbol_string)> 0)
                    then add_a_file_name(l_symbol_string);
                end;
              end; // analyse_the_uses_list

            var l_id, l_upper_id,
                l_previous_upper_id: String;

            begin // analyze_project_uses_list
              display('> analyze_project_uses_list');

              // -- here has read "UNIT xxx ;"
              skip_until_id('USES');

              if UpperCase(l_symbol_string)= 'USES'
                then analyse_the_uses_list;

              // -- here has read ';' of USES

              display('< analyze_project_uses_list');
            end; // analyze_project_uses_list

          begin // analyze_project
            analyze_project_name;
            analyze_project_uses_list;
          end; // analyze_project

        procedure analyze_unit;

          procedure analyze_unit_name;
            var l_unit_name: String;
            begin // analyze_unit_name
              display('> analyze_unit_name');

              // skip UNIT
              f_next_id;

              l_unit_name:= f_next_id;

              // skip ;
              read_symbol;

              display('< analyze_unit_name '+ l_unit_name+ '<');
            end; // analyze_unit_name

          procedure analyze_interface;

            procedure analyse_interface_uses_list;
              begin
                while not f_end_of_text and (l_symbol_string<> ';') do
                begin
                  read_symbol;
                  if (l_symbol_type= e_identifier)
                      and (Length(l_symbol_string)> 0)
                    then begin
                        add_a_file_name(l_symbol_string);
                      end;
                end;
              end; // analyse_interface_uses_list

            var l_id, l_upper_id,
                l_previous_upper_id: String;

            begin // analyze_interface
              display('> analyze_interface');

              // -- here has read "UNIT xxx ;"
              skip_until_id('INTERFACE');

              if UpperCase(f_next_id)= 'USES'
                then analyse_interface_uses_list;

              // -- here has read ';' of USES

              while not f_end_of_text and (UpperCase(f_next_id)<> 'IMPLEMENTATION') do ;

              display('< analyze_interface');
            end; // analyze_interface

          procedure analyze_implementation;

            procedure analyse_implementation_uses_list;
              begin
                while not f_end_of_text and (l_symbol_string<> ';') do
                begin
                  read_symbol;
                  if (l_symbol_type= e_identifier)
                      and (Length(l_symbol_string)> 0)
                    then begin
                        add_a_file_name(l_symbol_string);
                      end;
                end;
              end; // analyse_implementation_uses_list

            begin // analyze_implementation
              display('> analyze_implementation');

              // -- here has read 'IMPLEMENTATION'

              if UpperCase(f_next_id)= 'USES'
                then analyse_implementation_uses_list;

              display('< analyze_implementation');
            end; // analyze_implementation

          begin // analyze_unit
            analyze_unit_name;
            analyze_interface;
            analyze_implementation;
          end; // analyze_unit

        begin // analyze
          display('analyze '+ p_unit_name);

          if f_is_dpr(p_unit_name)
            then analyze_project
            else analyze_unit;
        end; // analyze

      var l_unit_name: String;

      procedure display_use_name;
        var l_c_use: c_use;
        begin
          with f_c_complete_file_name('name', p_unit_name) do
          begin
            l_unit_name:= m_name;

            l_c_use:= p_c_uses_list.f_c_use_by_name(m_name);
            if l_c_use<> nil
              then display(l_c_use.f_display_use);

            Free;
          end;
        end; // display_use_name

      begin // parse
        if f_load
          then begin
              display_use_name;

              analyze;

              // -- update the path of the analyzed file, and update "m_analyzed"
              p_c_uses_list.update_uses_list_path(l_unit_name, ExtractFilePath(p_unit_name));
              p_c_uses_list.display_uses_list;
            end // load ok
          else display('*** could not open '+ m_c_file_name.m_complete_name
              + ' '+ IntToStr(m_io_result));
      end; // parse

    destructor c_find_uses.Destroy;
      begin
        Inherited Destroy;
      end; // Destroy

    begin // u_c_find_uses_
    end. // u_c_find_uses




